<?php

declare(strict_types=1);

namespace Drupal\field_data\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\PagerSelectExtender;
use Drupal\Core\Database\Query\TableSortExtender;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Form\OptGroup;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\field\FieldConfigInterface;
use Drupal\field\FieldStorageConfigInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\StreamedResponse;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Returns responses for Field Data routes.
 */
class FieldDataController extends ControllerBase {

  /**
   * The current request.
   */
  protected Request $request;

  /**
   * The FieldController constructor.
   */
  public function __construct(
    protected Connection $database,
    protected EntityFieldManagerInterface $entityFieldManager,
  ) {}

  /**
   * Displays entity type and field data from the database.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return array
   *   A renderable containing entity type and field data from the database.
   */
  public function index(Request $request): array {
    $this->checkRequest($request);

    $this->request = $request;

    $entity_type = $request->query->get('entity_type');
    $field = $request->query->get('field');
    $bundle = $request->query->get('bundle');

    if ($entity_type && $field) {
      return $this->buildField($entity_type, $field, $bundle);
    }
    elseif ($entity_type && $bundle) {
      return $this->buildBundleFields($entity_type, $bundle);
    }
    elseif ($entity_type) {
      return $this->buildFields($entity_type);
    }
    else {
      return $this->buildEntityTypes();
    }
  }

  /**
   * Generates a CSV file from the requested entity data and returns it as a streamed response.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current HTTP request object containing the query parameters:
   *   - entity_type: The machine name of the entity type.
   *   - field: The field name to include in the CSV.
   *   - bundle (optional): The bundle machine name to filter the results.
   *
   * @return \Symfony\Component\HttpFoundation\StreamedResponse
   *   A streamed response containing the CSV file with the entity data.
   */
  public function csv(Request $request): StreamedResponse {
    $this->checkRequest($request);

    $entity_type = $request->query->get('entity_type');
    $field_name = $request->query->get('field');
    $bundle = $request->query->get('bundle');

    // Table mapping.
    /** @var \Drupal\Core\Entity\Sql\SqlContentEntityStorage $entity_storage */
    $entity_storage = $this->entityTypeManager()->getStorage($entity_type);
    $table_mapping = $entity_storage->getTableMapping();
    $table_name = $table_mapping->getFieldTableName($field_name);
    $all_columns = $table_mapping->getAllColumns($table_name);

    // Headers.
    $header = $all_columns;

    // Select table results.
    $select = $this->database->select($table_name, 't');
    $select->fields('t', $all_columns);
    if ($bundle) {
      $select->condition('bundle', $bundle);
    }
    // Add sortable header.
    $select = $select
      ->extend(TableSortExtender::class);
    // @phpstan-ignore-next-line
    $select->orderByHeader($header);
    // @phpstan-ignore-next-line
    $result = $select->execute();

    // Rows.
    $rows = [];
    while ($record = $result->fetchAssoc()) {
      $rows[] = $record;
    }

    // Response.
    $response = new StreamedResponse(function () use ($header, $rows): void {
      $handle = fopen('php://output', 'r+');
      fputcsv($handle, $header);
      foreach ($rows as $row) {
        fputcsv($handle, $row);
      }
      fclose($handle);
    });

    $parts = array_filter([
      $request->query->get('entity_type'),
      $request->query->get('bundle'),
      $request->query->get('field'),
    ]);
    $response->headers->set('Content-Type', 'application/force-download; charset=UTF-8');
    $response->headers->set('Content-Encoding', 'UTF-8');
    $response->headers->set('Content-Disposition', 'attachment; filename="' . implode('-', $parts) . '.csv"');
    return $response;
  }

  /**
   * Build browse entity types page.
   *
   * @return array[]
   *   A renderable array containing the browse entity types page.
   */
  protected function buildEntityTypes(): array {
    /** @var \Drupal\field\FieldStorageConfigInterface[] $field_config_storages */
    $field_config_storages = $this->entityTypeManager()->getStorage('field_storage_config')->loadMultiple();
    $entity_types = [];
    foreach ($field_config_storages as $field_config_storage) {
      $entity_type_id = $field_config_storage->getTargetEntityTypeId();
      $entity_types[$entity_type_id] = $entity_type_id;
    }
    asort($entity_types);

    $content = [];
    foreach ($entity_types as $entity_type_id) {
      $content[] = [
        'title' => $entity_type_id,
        'url' => $this->url($entity_type_id),
        'options' => [],
      ];
    }

    return [
      'admin_block_content' => [
        '#theme' => 'admin_block_content',
        '#content' => $content,
      ],
    ];
  }

  /**
   * Build a fields configuration page for the given entity type.
   *
   * @param string $entity_type
   *   The ID of the entity type for which to build the fields.
   *
   * @return array
   *   A renderable array containing the fields configuration page.
   */
  protected function buildFields(string $entity_type): array {
    $field_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type);

    // Header.
    $header = [];
    $header[] = ['data' => $this->t('Name'), 'width' => '20%'];
    $header[] = ['data' => $this->t('Type'), 'width' => '15%'];
    $header[] = ['data' => $this->t('Cardinality'), 'width' => '5%'];
    $header[] = ['data' => $this->t('Used in'), 'width' => '20%'];
    $header[] = ['data' => $this->t('Options'), 'width' => '20%'];
    $header[] = ['data' => $this->t('Number of records'), 'width' => '20%'];

    $rows = [];
    foreach ($field_definitions as $field_definition) {
      if (!$field_definition instanceof FieldStorageConfigInterface) {
        continue;
      }

      $field_name = $field_definition->getName();

      $row = [];
      $row[] = [
        'data' => [
          '#type' => 'link',
          '#title' => $field_definition->getName(),
          '#url' => $this->url(
            entity_type: $entity_type,
            field_name: $field_name,
          ),
        ],
      ];
      $row[] = $field_definition->getType();

      $row[] = ($field_definition->getCardinality() === -1)
        ? $this->t('Unlimited')
        : $field_definition->getCardinality();

      $used_id = [];
      $field_bundle = NULL;
      $field_bundles = $field_definition->getBundles();
      foreach ($field_bundles as $field_bundle) {
        $used_id[$field_bundle] = [
          '#type' => 'link',
          '#title' => $field_bundle,
          '#url' => $this->url(
            entity_type: $entity_type,
            bundle: $field_bundle,
          ),
          '#suffix' => ', ',
        ];
      }
      unset($used_id[$field_bundle]['#suffix']);
      $row[] = ['data' => $used_id];

      $row[] = $this->buildOptions($field_definition);

      $counts = $this->getFieldCounts($entity_type, $field_name);
      $totals = [];
      $totals[] = [
        '#type' => 'link',
        '#title' => $this->t('All'),
        '#url' => $this->url(
          entity_type: $entity_type,
          field_name: $field_name,
        ),
        '#suffix' => ' [' . array_sum($counts) . ']',
      ];
      foreach ($field_bundles as $field_bundle) {
        $totals[] = [
          '#type' => 'link',
          '#title' => $field_bundle,
          '#url' => $this->url(
            entity_type: $entity_type,
            bundle: $field_bundle,
            field_name: $field_name,
          ),
          '#suffix' => ' [' . ($counts[$field_bundle] ?? 0) . ']',
        ];
      }
      $row[] = [
        'data' => [
          '#type' => 'details',
          '#title' => $this->formatPlural(array_sum($counts), '1 record', '@count records'),
          'bundles' => [
            '#theme' => 'item_list',
            '#items' => $totals,
          ],
        ],
      ];

      $rows[] = $row;
    }

    return [
      'info' => [
        '#markup' => $this->formatPlural(count($rows), '1 field', '@count fields'),
        '#prefix' => '<p>',
        '#suffix' => '</p>',
      ],
      'table' => [
        '#theme' => 'table',
        '#header' => $header,
        '#rows' => $rows,
      ],
    ];
  }

  /**
   * Build a fields configuration page for the given entity type.
   *
   * @param string $entity_type
   *   The ID of the entity type for which to build the fields.
   * @param string $bundle
   *   The ID of the bundle to filter by, or NULL to include all bundles.
   *
   * @return array
   *   A renderable array containing the fields configuration page.
   */
  protected function buildBundleFields(string $entity_type, string $bundle): array {
    $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type, $bundle);

    // Header.
    $header = [];
    $header[] = ['data' => $this->t('Name'), 'width' => '20%'];
    $header[] = ['data' => $this->t('Type'), 'width' => '10%'];
    $header[] = ['data' => $this->t('Cardinality'), 'width' => '5%'];
    $header[] = ['data' => $this->t('Required'), 'width' => '5%'];
    $header[] = ['data' => $this->t('Options'), 'width' => '20%'];
    $header[] = ['data' => $this->t('Number of records'), 'width' => '20%'];

    $rows = [];
    foreach ($field_definitions as $field_definition) {
      if (!$field_definition instanceof FieldConfigInterface) {
        continue;
      }

      $field_name = $field_definition->getName();
      $field_storage = $field_definition->getFieldStorageDefinition();

      $row = [];
      $row[] = [
        'data' => [
          '#type' => 'link',
          '#title' => $field_definition->getName(),
          '#url' => $this->url(
            entity_type: $entity_type,
            bundle: $bundle,
            field_name: $field_name,
          ),
        ],
      ];

      $row[] = $field_definition->getType();

      $row[] = ($field_storage->getCardinality() === -1)
        ? $this->t('Unlimited')
        : $field_storage->getCardinality();

      $row[] = ($field_definition->isRequired())
        ? $this->t('Yes')
        : $this->t('No');

      $row[] = $this->buildOptions($field_storage);

      $counts = $this->getFieldCounts($entity_type, $field_name);
      $row[] = $this->formatPlural(($counts[$bundle] ?? 0), '1 record', '@count records');

      $rows[] = $row;
    }

    return [
      'info' => [
        '#markup' => $this->formatPlural(count($rows), '1 field', '@count fields'),
        '#prefix' => '<p>',
        '#suffix' => '</p>',
      ],
      'table' => [
        '#theme' => 'table',
        '#header' => $header,
        '#rows' => $rows,
      ],
    ];
  }

  /**
   * Builds a table render array for a specific field in an entity type.
   *
   * @param string $entity_type
   *   The ID of the entity type.
   * @param string $field_name
   *   The name of the field to display.
   * @param string|null $bundle
   *   (optional) The ID of the bundle_name to filter by, or NULL to include all bundles.
   *
   * @return array
   *   A renderable array containing the table of field data.
   */
  protected function buildField(string $entity_type, string $field_name, ?string $bundle = NULL): array {
    // Table mapping.
    /** @var \Drupal\Core\Entity\Sql\SqlContentEntityStorage $entity_storage */
    $entity_storage = $this->entityTypeManager()->getStorage($entity_type);
    $table_mapping = $entity_storage->getTableMapping();
    $table_name = $table_mapping->getFieldTableName($field_name);
    $all_columns = $table_mapping->getAllColumns($table_name);
    $field_columns = $table_mapping->getColumnNames($field_name);

    // Get number of records.
    $counts = $this->getFieldCounts($entity_type, $field_name);
    $count = ($bundle)
      ? $counts[$bundle] ?? 0
      : array_sum($counts);

    // Headers.
    $header = [];
    foreach ($all_columns as $column_name) {
      $header[$column_name] = [
        'data' => $column_name,
        'field' => $column_name,
        'specifier' => $column_name,
      ];
    }

    // Select table results.
    $select = $this->database->select($table_name, 't');
    $select->fields('t', $all_columns);
    if ($bundle) {
      $select->condition('bundle', $bundle);
    }

    // Display only published and/or default translations.
    $published = $this->config('field_data.settings')->get('published');
    $default_langcode = $this->config('field_data.settings')->get('default_langcode');
    if ($published || $default_langcode) {
      $entity_definition = $this->entityTypeManager()->getDefinition($entity_type);

      $data_table = $entity_definition->getDataTable();

      $id_key = $entity_definition->getKey('id');
      $revision_key = $entity_definition->getKey('revision');
      $langcode_key = $entity_definition->getKey('langcode');

      $select->innerJoin($data_table, 'd', "t.entity_id = d.{$id_key} AND t.revision_id = d.{$revision_key}  AND t.langcode = d.{$langcode_key}");
      $status_key = $entity_definition->getKey('status');
      if ($published && $status_key) {
        $select->condition("d.{$status_key}", 1);
      }
      if ($default_langcode && $entity_definition->isTranslatable()) {
        $select->condition('d.default_langcode', 1);
      }
    }

    // Add pager and sortable header.
    $select = $select
      ->extend(PagerSelectExtender::class)
      ->extend(TableSortExtender::class);
    // @phpstan-ignore-next-line
    $select->limit(500)->orderByHeader($header);
    // @phpstan-ignore-next-line
    $result = $select->execute();

    // Rows.
    $rows = [];
    while ($record = $result->fetchAssoc()) {
      // Add <br> to field values.
      foreach ($field_columns as $column_name) {
        $value = $record[$column_name];
        if (is_string($value) && str_contains($value, "\n")) {
          $record[$column_name] = ['data' => ['#markup' => nl2br(htmlentities($value))]];
        }
      }

      if (empty($bundle)) {
        $record['bundle'] = [
          'data' => [
            '#type' => 'link',
            '#title' => $record['bundle'],
            '#url' => $this->url(
              entity_type: $entity_type,
              bundle: (string) $record['bundle'],
              field_name: $field_name,
            ),
          ],
        ];
      }

      $record['entity_id'] = [
        'data' => [
          '#type' => 'link',
          '#title' => $record['entity_id'],
          '#url' => Url::fromRoute(
            'entity.' . $entity_type . '.edit_form',
            [$entity_type => $record['entity_id']]
          ),
        ],
      ];

      $rows[] = $record;
    }

    $t_args = [
      '%field' => implode('.', array_filter([$entity_type, $bundle, $field_name])),
    ];
    return [
      'info' => [
        'records' => [
          '#markup' => $this->formatPlural($count, 'Displaying 1 record for the %field field', 'Displaying @count records for the %field field', $t_args),
        ],
        'all' => ($bundle) ? [
          '#type' => 'link',
          '#title' => $this->t('View all'),
          '#url' => $this->url(
            entity_type: $entity_type,
            field_name: $field_name,
          ),
          '#prefix' => ' [',
          '#suffix' => ']',
        ] : [],
        '#prefix' => '<p>',
        '#suffix' => '</p>',
      ],
      'table' => [
        '#theme' => 'table',
        '#header' => $header,
        '#rows' => $rows,
      ],
      'pager' => [
        '#type' => 'pager',
      ],
      'download' => [
        '#type' => 'link',
        '#title' => $this->t('Download CSV'),
        '#url' => Url::fromRoute(
          'field_data.csv',
          [],
          ['query' => $this->request->query->all()]
        ),
        '#attributes' => ['class' => ['button', 'button--small']],
        '#prefix' => '<p>',
        '#suffix' => '</p>',
      ],
    ];
  }

  /**
   * Returns the page title for migrations data.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The translated page title.
   */
  public function title(Request $request): TranslatableMarkup {
    $parts = array_filter([
      $request->query->get('entity_type'),
      $request->query->get('bundle'),
      $request->query->get('field'),
    ]);

    return ($parts)
      ? $this->t('Field data: @title', ['@title' => implode('.', $parts)])
      : $this->t('Field data');
  }

  /**
   * Validates the incoming request for entity type and field.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The HTTP request object containing query parameters.
   */
  protected function checkRequest(Request $request): void {
    $properties = array_filter([
      'entity_type' => $request->query->get('entity_type'),
      'bundle' => $request->query->get('bundle'),
      'field_name' => $request->query->get('field'),
    ]);
    if (empty($properties)) {
      return;
    }

    $field_configs = $this->entityTypeManager()
      ->getStorage('field_config')
      ->loadByProperties($properties);
    if (empty($field_configs)) {
      throw new NotFoundHttpException();
    }
  }

  /**
   * Generates a URL based on entity type, bundle, and field name.
   *
   * @param string $entity_type
   *   The ID of the entity type.
   * @param string|null $bundle
   *   The bundle name, or NULL for no bundle.
   * @param string|null $field_name
   *   The field name, or NULL for no specific field.
   *
   * @return \Drupal\Core\Url
   *   The generated URL object.
   */
  protected function url(string $entity_type, ?string $bundle = '', ?string $field_name = ''): Url {
    $query = [
      'entity_type' => $entity_type,
      'bundle' => $bundle,
      'field' => $field_name,
    ];
    return Url::fromRoute(
      'field_data.report',
      [],
      ['query' => array_filter($query)]
    );
  }

  /**
   * Retrieves field counts grouped by bundle for a given entity type and field.
   *
   * @param string $entity_type
   *   The entity type machine name.
   * @param string $field_name
   *   The field name for which to retrieve counts.
   *
   * @return array
   *   An associative array where the keys are bundle names and the values are the counts of the field.
   */
  protected function getFieldCounts(string $entity_type, string $field_name): array {
    /** @var \Drupal\Core\Entity\Sql\SqlContentEntityStorage $entity_storage */
    $entity_storage = $this->entityTypeManager()->getStorage($entity_type);
    $table_mapping = $entity_storage->getTableMapping();

    $table_name = $table_mapping->getFieldTableName($field_name);
    $select = $this->database->select($table_name, 't');
    $select->addField('t', 'bundle');
    $select->addExpression('COUNT(t.bundle)', 'count');
    $select->groupBy('t.bundle');
    return $select->execute()->fetchAllKeyed();
  }

  /**
   * Builds options for the provided field storage configuration.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $field_storage
   *   The field storage configuration to build options for.
   *
   * @return array
   *   A renderable array containing the options details, or an empty array if no options are available.
   */
  protected function buildOptions(FieldStorageDefinitionInterface $field_storage): array {
    $options = options_allowed_values($field_storage);
    if ($options) {
      $options = OptGroup::flattenOptions($options);
      foreach ($options as $key => &$value) {
        if ($key !== $value) {
          $value .= ' (' . $key . ')';
        }
      }
      return [
        'data' => [
          '#type' => 'details',
          '#title' => $this->formatPlural(count($options), '1 option', '@count options'),
          'bundles' => [
            '#theme' => 'item_list',
            '#items' => $options,
          ],
        ],
      ];
    }
    else {
      return [];
    }
  }

}
